#include<alexandria.h>
#include<stdio.h>

int main(int argc, char **argv) {
    printf("Ahh, another visitor. Stay a while.\n");
    printf("You enter the library.\n\n");
    alexandria_visit();
    printf("\nYou decided not to stay forever.\n");
    return 0;
}
