#include "rejected.h"

void say(void) {
    printf("You are standing outside the Great Library of Alexandria.\n");
    printf("You decide to go inside.\n\n");
    alexandria_visit();
    printf("The librarian tells you it's time to leave\n");
}
