#!/usr/bin/env python3
# SPDX-license-identifer: Apache-2.0
# Copyright © 2021 Intel Corporation

import argparse
import shutil


def main() -> None:
    parser = argparse.ArgumentParser()
    parser.add_argument('input')
    parser.add_argument('output')
    args = parser.parse_args()

    shutil.copy2(args.input, args.output)


if __name__ == "__main__":
    main()
