#include <zlib.h>
#include <stdio.h>
#include <string.h>

int main(void) {
    void * something = deflate;
    if(strcmp(ZLIB_VERSION, FOUND_ZLIB) != 0) {
        printf("Meson found '%s' but zlib is '%s'\n", FOUND_ZLIB, ZLIB_VERSION);
        return 2;
    }
    if(something != 0)
        return 0;
    printf("Couldn't find 'deflate'\n");
    return 1;
}
