#include<up_down.h>
#include<stdio.h>

int main(int argc, char **argv) {
    if(argc == 42) {
        printf("Very sneaky, %s\n", argv[0]);
    }
#ifdef UP_IS_DOWN
    return 0;
#else
    return 1;
#endif
}
